@@ -78,7 +78,9 @@ class SSHConnection:
         hkeys = ssh.get_host_keys()
         hkeys.clear()
         ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
-        ssh.connect(host, username=user, key_filename=privkey, allow_agent=False, look_for_keys=False )
+        ssh.connect(host, username=user, key_filename=privkey, allow_agent=False, look_for_keys=False)
+        transport = ssh.get_transport()
+        transport.set_keepalive(30)
         return ssh
     
     def run(self, privkey, host, user, cmd):
@@ -185,22 +187,28 @@ class Experiment:
     Base class for all types of cloud implementations. This is what defines the experiment API.
     """
     
-    def __init__(self):
+    def __init__(self, name = None):
         """
         Constructor for a new experiment - this will set up ~/.precip and ssh keys if they
         do not already exist in a way that you can use precip from multiple machines or 
         accounts at the same time. 
         """
         
+        if name is None:
+            self._name = str(int(time.time()))
+        else:
+            self._name = name
+
         self._instances = []
         
         self._conf_dir = os.path.join(os.environ["HOME"], ".precip")
-        uid = self._get_account_id()
         
         # checking/creating conf directory
         if not os.path.exists(self._conf_dir):
             os.makedirs(self._conf_dir)
         
+        uid = self._get_account_id()
+        
         # ssh keys setup 
         self._ssh_pubkey = os.path.join(self._conf_dir, "precip_"+uid+".pub")
         self._ssh_privkey = os.path.join(self._conf_dir, "precip_"+uid)
@@ -214,6 +222,12 @@ class Experiment:
             if rc != 0:
                 raise ExperimentException("Command '%s' failed with error code %s" % (cmd, rc))
 
+    def __del__(self):
+        """
+        Deprovision all instances
+        """
+        self.deprovision([])
+
     def _instance_subset(self, tags):
         """
         Returns the subset of instances matching the tags
@@ -275,7 +289,7 @@ class Experiment:
             f.close()
         return uid   
         
-    def provision(self, image_id, instance_type='m1.small', count=1, tags=None):
+    def provision(self, image_id, instance_type='m1.small', count=1, tags=None, boot_timeout=600, boot_max_tries=3):
         """
         Provision a new instance. Note that this method starts the provisioning cycle, but does not
         block for the instance to finish booting - for that, see wait()
@@ -285,10 +299,12 @@ class Experiment:
         :param count: Number of instances to provision. The default is 1.
         :param tags: Tags to add to the instance - this is important as tags are used throughout the API 
                      to find and manipulate instances
+        :param boot_timeout: The amount of allowed time in seconds for an instance to boot
+        :param boot_max_tries: The number of tries an instance is given to successfully boot
         """                                                                   
         pass
     
-    def wait(self, tags=[], timeput=300, maxretrycount=3):
+    def wait(self, tags=[]):
         """
         Barrier for all currently instances to finish booting and be accessible via external addresses.
         
@@ -378,13 +394,15 @@ class Experiment:
         for i in self._instance_subset(tags):
             ssh.put(self._ssh_privkey, i.pub_addr, user, local_path, remote_path)
     
-    def run(self, tags, cmd, user="root", check_exit_code=True):
+    def run(self, tags, cmd, user="root", check_exit_code=True, output_base_name=None):
         """
         Runs a command on set of instances matching the tags given.
         
         :param tags: set of tags to match against
         :param cmd: command to run
         :param user: the user to run the command as
+        :param check_exit_code: if true, non-zero exit codes will be considered fatal
+        :param output_base_name: redirects output to a file instead of stdout
         """
         exit_code_list = []
         out_list = []
@@ -399,10 +417,30 @@ class Experiment:
                 exit_code, out, err = ssh.run(self._ssh_privkey, i.pub_addr, user, cmd)
             except Exception, e:
                 raise ExperimentException("Error running ssh command", e)
+
             if len(out) > 0:
-                logger.info("  stdout: %s" % out)
+                if output_base_name is not None:
+                    fname = "%s.%s.stdout" %(output_base_name, i.id)
+                    try:
+                        f = open(fname, 'w')
+                        f.write(out)
+                        f.close()
+                    except Exception, e:
+                        raise ExperimentException("Unable to write to " + fname, e)
+                else:
+                    logger.info("  stdout: %s" % out)
+
             if len(err) > 0:
-                logger.info("  stderr: %s" % err)
+                if output_base_name is not None:
+                    fname = "%s.%s.stderr" %(output_base_name, i.id)
+                    try:
+                        f = open(fname, 'w')
+                        f.write(err)
+                        f.close()
+                    except Exception, e:
+                        raise ExperimentException("Unable to write to " + fname, e)
+                else:
+                    logger.info("  stderr: %s" % err)
             
             exit_code_list.append(exit_code)
             out_list.append(out)
@@ -435,10 +473,7 @@ class Experiment:
 
 class EC2Experiment(Experiment):
     
-    _image_id = None
-    _instance_type = None
-     
-    def __init__(self, region, endpoint, access_key, secret_key):
+    def __init__(self, region, endpoint, access_key, secret_key, name = None):
         """
         Initializes an EC2 experiment
         
@@ -447,7 +482,7 @@ class EC2Experiment(Experiment):
         :param access_keys: Amazon EC2 access key
         :param secret_keys: Amazon EC2 secret key
         """        
-        Experiment.__init__(self)
+        Experiment.__init__(self, name = name)
     
         self._region = region
         self._endpoint = endpoint
@@ -477,7 +512,7 @@ class EC2Experiment(Experiment):
         if not r:
             raise ExperimentException("Unable to parse endpoint: %s" % (self._endpoint))
 
-        # Parse successfu
+        # Parse successful
         proto = r.group(2)
         host = r.group(3)
         port = r.group(4)
@@ -525,7 +560,7 @@ class EC2Experiment(Experiment):
     def _ssh_keys_setup(self):
         
         """
-   Makes sure we have our experiment keypair registered
+        Makes sure we have our experiment keypair registered
         """
         uid = self._get_account_id()
         keypairs = None
@@ -537,7 +572,7 @@ class EC2Experiment(Experiment):
             # not found on eucalyptus
             pass
         except EC2ResponseError, e:
-            if e.error_code in ["InvalidKeyPair.NotFound", "EC2APIError"]:
+            if e.error_code in ["InvalidKeyPair.NotFound", "KeypairNotFound", "EC2APIError"]:
                 keypairs = None
             else:
                 raise ExperimentException("Unable to query for key pair", e)
@@ -550,7 +585,6 @@ class EC2Experiment(Experiment):
             f.close()
             self._conn.import_key_pair("precip_"+uid, contents) 
               
-                   
     def _security_groups_setup(self):
         """
         Sets up the default security group
@@ -574,7 +608,51 @@ class EC2Experiment(Experiment):
                 logger.warn("Security group seems to be broken - disabling support")
                 self._security_groups_support = False
                 pass
-    
+
+    def _start_instance(self, image_id, instance_type, ebs_size):
+        """
+        Creates a new instance
+        
+        :param instance: the instance to check
+        :return: the instance Boto object
+        """
+        uid = self._get_account_id()
+        boto_instance = None
+        try:
+            # block device maps is only needed if the user wants to specify ebs_size
+            block_device_map = None
+            if ebs_size is not None:
+                dev_sda1 = boto.ec2.blockdevicemapping.EBSBlockDeviceType(delete_on_termination = True)
+                dev_sda1.size = int(ebs_size)
+                block_device_map = boto.ec2.blockdevicemapping.BlockDeviceMapping()
+                block_device_map['/dev/sda1'] = dev_sda1
+            #else:
+            #    block_device_map = self._conn.get_image_attribute(image_id, 
+            #                                                      attribute = 'blockDeviceMapping')
+
+            # create a boto image object from the image id          
+            image_obj = self._conn.get_image(image_id)
+            if image_obj is None:
+                raise ExperimentException("Image %s does not exist" %(image_id))
+
+            if self._security_groups_support:
+                res = image_obj.run(instance_type = instance_type,
+                                    key_name = "precip_"+uid,
+                                    instance_initiated_shutdown_behavior = "terminate",
+                                    block_device_map = block_device_map,
+                                    security_groups = ["precip"])
+            else:
+                res = image_obj.run(instance_type = instance_type,
+                                    key_name = "precip_"+uid,
+                                    instance_initiated_shutdown_behavior = "terminate",
+                                    block_device_map = block_device_map)
+            boto_instance = res.instances[0]
+
+            logger.info("Started instance %s, type %s" % (boto_instance.id, instance_type))        
+        except Exception as e:
+            raise ExperimentException("Unable to provision a new instance", e)
+        return res.instances[0]
+
     def _finish_instanciation(self, instance):
         """
         Finishes booting and bootstraps an instace
@@ -582,6 +660,8 @@ class EC2Experiment(Experiment):
         :param instance: the instance to check
         :return: True if the instance is ready, otherwise False
         """
+        instance_record = []
+        count_ing = 0
         # check if we are already done
         if instance.is_fully_instanciated:
             return True
@@ -590,6 +670,11 @@ class EC2Experiment(Experiment):
         ec2inst = instance.ec2_instance
         ec2inst.update()
         
+        if ec2inst.state == "error":
+            logger.debug("Instance %s state is 'error - scheduling for possible retry" %instance.id)
+            instance.boot_timeout = 0
+            return False
+        
         if ec2inst.state != "pending" and ec2inst.state != "running":
             
             #raise ExperimentException("Unexpected instance state for instance %s: %s" % (instance.id, ec2inst.state))
@@ -599,6 +684,15 @@ class EC2Experiment(Experiment):
         if ec2inst.state == "pending":
             logger.debug("Instance %s is still pending" % instance.id)
             return False
+         
+        #if ec2inst.public_dns_name == ec2inst.private_dns_name:
+        #    # we did not get a public address assigned to us, do it now
+        #    logger.debug("Requesting a public IP address")
+        #    addr = self._conn.allocate_address()
+        #    #self._conn.associate_address(instance_id = instance.id, public_ip = addr)
+        #    logger.debug("Got public ip: %s" %(addr))
+        #    ec2inst.use_ip(addr)
+        #    return False
     
         if not self._is_valid_hostaddr(ec2inst.public_dns_name):
             logger.debug("Waiting for instance %s to boot and be assigned a public IP address" % instance.id)
@@ -634,12 +728,26 @@ class EC2Experiment(Experiment):
             raise ExperimentException("Bootstrap script exited with error %d" % exit_code)
         
         logger.info("Instance %s has booted, public hostname: %s" % (instance.id, ec2inst.public_dns_name))
+
+        # add our tags
+        try:
+            ec2inst.add_tag("Name", "PRECIP - " + self._name)
+            # can only add 10 on EC2
+            tag_count = 1
+            for t in instance.tags:
+                if tag_count >= 10:
+                    break
+                ec2inst.add_tag(t, "1")
+                tag_count += 1
+        except Exception, e:
+            # ignore - the infrastructure might not support user tags
+            pass
         
         instance.add_tag(instance.pub_addr)
         instance.is_fully_instanciated = True
         return True
-    
-    def _retry(self, instance, image_id, instance_type, tags):
+
+    def _retry(self, instance):
         
         """
         In case of reaching timeout for an instance, retry will terminate the previous instance and 
@@ -660,30 +768,18 @@ class EC2Experiment(Experiment):
             self._conn.terminate_instances(instance_ids=[instance.id])
         except AttributeError as e:
             logger.warn("Terminating issued an attribute warning")
-        self._instances.remove(instance)
-
-        try:
-            image_obj = self._conn.get_image(image_id)    
-            if self._security_groups_support:
-                res = image_obj.run(instance_type=instance_type, key_name="precip_"+uid, security_groups=["precip"])
-            else:
-                res = image_obj.run(instance_type=instance_type, key_name="precip_"+uid)
-            logger.info("Started a new instance %s, type %s" % (res.instances[0].id, instance_type))
         except Exception as e:
-            raise ExperimentException("Unable to provision a new instance", e)
-                
-        instance = Instance(res.instances[0].id)
-        instance.ec2_instance = res.instances[0]
-            
-        instance.add_tag("precip")
-        instance.add_tag(instance.id)
-        for t in tags:
-            instance.add_tag(t)
-            
-        self._instances.append(instance)
+            logger.warn("Ignoring error while terminating instance", e)
+
+        boto_inst = self._start_instance(instance.image_id, instance.instance_type, instance.ebs_size)
+        instance.id = boto_inst.id
+        instance.ec2_instance = boto_inst
+        instance.num_starts = instance.num_starts + 1
+        instance.boot_time = int(time.time())
 
             
-    def provision(self, image_id, instance_type='m1.small', count=1, tags=None):                                                                   
+    def provision(self, image_id, instance_type='m1.small', count=1, ebs_size=None, tags=None,
+                  boot_timeout=600, boot_max_tries=3):
         """
         Provision a new instance. Note that this method starts the provisioning cycle, but does not
         block for the instance to finish booting - for that, see wait()
@@ -693,29 +789,28 @@ class EC2Experiment(Experiment):
         :param count: Number of instances to provision. The default is 1.
         :param tags: Tags to add to the instance - this is important as tags are used throughout the API 
                      to find and manipulate instances
+        :param boot_timeout: The amount of allowed time in seconds for an instance to boot
+        :param boot_max_tries: The number of tries an instance is given to successfully boot
         """   
         
         uid = self._get_account_id()
-        self._instance_type = instance_type
-        self._image_id = image_id
         
         self._get_connection()
                
         for i in range(count):
-            try:
-                # create a boto image object from the image id
-                image_obj = self._conn.get_image(image_id)
+            boto_inst = self._start_instance(image_id, instance_type, ebs_size)
 
-                if self._security_groups_support:
-                    res = image_obj.run(instance_type=instance_type, key_name="precip_"+uid, security_groups=["precip"])
-                else:
-                    res = image_obj.run(instance_type=instance_type, key_name="precip_"+uid)
-                logger.info("Started instance %s, type %s" % (res.instances[0].id, instance_type))        
-            except Exception as e:
-                raise ExperimentException("Unable to provision a new instance", e)
-        
-            instance = Instance(res.instances[0].id)
-            instance.ec2_instance = res.instances[0]
+            instance = Instance(boto_inst.id)
+            instance.ec2_instance = boto_inst
+
+            # keep track of parameters - we might need them for restarts later
+            instance.num_starts = 1
+            instance.boot_time = int(time.time())
+            instance.boot_timeout = boot_timeout
+            instance.boot_max_tries = 3
+            instance.image_id = image_id
+            instance.instance_type = instance_type
+            instance.ebs_size = ebs_size
             
             # add basic tags
             instance.add_tag("precip")
@@ -724,8 +819,8 @@ class EC2Experiment(Experiment):
                 instance.add_tag(t)
             
             self._instances.append(instance)
-        
-    def wait(self, tags=[], timeout=300, maxretrycount=3):
+
+    def wait(self, tags=[]):
         """
         Barrier for all currently instances to finish booting and be accessible via external addresses.
         
@@ -733,33 +828,27 @@ class EC2Experiment(Experiment):
         :param timeout: maximum timeout to wait for instances to come up
         :param maxretrycount: The number of retries in case of failure of provisioning the instances 
         """
-        retrycount = maxretrycount
-        imageID = self._image_id
-        instanceType = self._instance_type
         
-        start_time = int(time.time())
         count_pending = -1
         while count_pending != 0:
             count_pending = 0
-            pending_set = []
+            current_time = int(time.time())
+
             for i in self._instance_subset(tags):
                 if not self._finish_instanciation(i):
                     count_pending += 1
-                    pending_set.append(i)
+
+                    # did the instance timeout?
+                    if current_time > i.boot_time + i.boot_timeout:
+                        logger.info("Timeout reached while waiting for instances to boot")
+                        if i.num_starts < i.boot_max_tries:
+                            self._retry(i)
+                        else:
+                            raise ExperimentException("Timeout reached while waiting for instances to boot")
+
             if count_pending > 0:
-                current_time = int(time.time())
-                if start_time + timeout < current_time:
-                    logger.info("Timeout reached while waiting for instances to boot")
-                    if retrycount > 1:
-                        for j in pending_set:
-                            self._retry(j, imageID, instanceType, tags)
-                        retrycount -= 1
-                        start_time = int(time.time())
-                    else:
-                        raise ExperimentException("Timeout reached while waiting for instances to boot")
-                logger.info("Waiting for %d instances to finish booting" % (count_pending))
+                logger.info("Still waiting for %d instances to finish booting" % (count_pending))
                 time.sleep(30)       
-    
             
     def deprovision(self, tags=[]):
         """
@@ -782,7 +871,7 @@ class OpenStackExperiment(EC2Experiment):
     A class defining an experiment running on top of OpenStack
     """
     
-    def __init__(self, endpoint, access_key, secret_key):
+    def __init__(self, endpoint, access_key, secret_key, name = None):
         """
         Initializes an OpenStack experiment
         
@@ -798,7 +887,7 @@ class EucalyptusExperiment(EC2Experiment):
     A class defining an experiment running on top of Eucalyptus
     """
     
-    def __init__(self, endpoint, access_key, secret_key):
+    def __init__(self, endpoint, access_key, secret_key, name = None):
         """
         Initializes an Eucalyptus experiment
         
@@ -814,7 +903,7 @@ class NimbusExperiment(EC2Experiment):
     A class defining an experiment running on top of Nimbus
     """
     
-    def __init__(self, endpoint, access_key, secret_key):
+    def __init__(self, endpoint, access_key, secret_key, name = None):
         """
         Initializes a Nimbus experiment
         
